use std::error::Error;
use std::fmt::Display;

use futures::{SinkExt, StreamExt};
use tokio_tungstenite::tungstenite::Message;

use crate::data::{BoatsLayout, BotType, GameRules};
use crate::human_player_ws::{ClientMessage, ServerMessage};
use crate::server::{AcceptInviteQuery, BotPlayQuery, CreateInviteQuery, PlayRandomQuery};

const PLAYER_NAME: &str = "Bot client";

#[derive(Default)]
pub enum RunMode {
    #[default]
    AgainstBot,
    CreateInvite,
    AcceptInvite {
        code: String,
    },
    Random,
}

#[derive(Debug, Eq, PartialEq)]
pub enum ClientEndResult {
    Finished {
        number_victories: usize,
        number_defeats: usize,
    },
    InvalidBoatsLayout,
    OpponentRejectedRematch,
    OpponentLeftGame,
    InvalidInviteCode,
}

pub struct BotClient {
    server: String,
    run_mode: RunMode,
    requested_rules: GameRules,
    layout: Option<BoatsLayout>,
    number_plays: usize,
    server_msg_callback: Option<Box<dyn FnMut(&ServerMessage)>>,
    play_as_bot_type: BotType,
}

impl BotClient {
    pub fn new<D>(server: D) -> Self
    where
        D: Display,
    {
        Self {
            server: server.to_string(),
            run_mode: RunMode::default(),
            requested_rules: GameRules::random_players_rules(),
            layout: None,
            number_plays: 1,
            server_msg_callback: None,
            play_as_bot_type: BotType::Random,
        }
    }

    pub fn with_run_mode(mut self, mode: RunMode) -> Self {
        self.run_mode = mode;
        self
    }

    pub fn with_rules(mut self, rules: GameRules) -> Self {
        self.requested_rules = rules;
        self
    }

    pub fn with_layout(mut self, layout: BoatsLayout) -> Self {
        self.layout = Some(layout);
        self
    }

    pub fn with_number_plays(mut self, number: usize) -> Self {
        self.number_plays = number;
        self
    }

    pub fn with_server_msg_callback<F>(mut self, cb: F) -> Self
    where
        F: FnMut(&ServerMessage) + 'static,
    {
        self.server_msg_callback = Some(Box::new(cb));
        self
    }

    pub fn with_play_as_bot_type(mut self, t: BotType) -> Self {
        self.play_as_bot_type = t;
        self
    }

    pub async fn run_client(&mut self) -> Result<ClientEndResult, Box<dyn Error>> {
        let mut remaining_games = self.number_plays;
        let mut number_victories = 0;
        let mut number_defeats = 0;

        let url = match &self.run_mode {
            RunMode::AgainstBot => {
                format!(
                    "{}/play/bot?{}",
                    self.server.replace("http", "ws"),
                    serde_urlencoded::to_string(&BotPlayQuery {
                        rules: self.requested_rules.clone(),
                        player_name: PLAYER_NAME.to_string()
                    })
                    .unwrap()
                )
            }
            RunMode::CreateInvite => {
                format!(
                    "{}/play/create_invite?{}",
                    self.server.replace("http", "ws"),
                    serde_urlencoded::to_string(&CreateInviteQuery {
                        rules: self.requested_rules.clone(),
                        player_name: PLAYER_NAME.to_string(),
                    })
                    .unwrap()
                )
            }
            RunMode::AcceptInvite { code } => {
                format!(
                    "{}/play/accept_invite?{}",
                    self.server.replace("http", "ws"),
                    serde_urlencoded::to_string(&AcceptInviteQuery {
                        code: code.to_string(),
                        player_name: PLAYER_NAME.to_string()
                    })
                    .unwrap()
                )
            }
            RunMode::Random => {
                format!(
                    "{}/play/random?{}",
                    self.server.replace("http", "ws"),
                    serde_urlencoded::to_string(&PlayRandomQuery {
                        player_name: PLAYER_NAME.to_string()
                    })
                    .unwrap()
                )
            }
        };
        log::debug!("Connecting to  {}...", url);
        let (mut socket, _) = match tokio_tungstenite::connect_async(url).await {
            Ok(s) => s,
            Err(e) => {
                log::error!("Failed to establish WebSocket connection! {:?}", e);
                return Err(Box::new(e));
            }
        };

        while let Some(chunk) = socket.next().await {
            let message = match chunk? {
                Message::Text(message) => {
                    log::trace!("TEXT message from server: {}", message);

                    let msg: ServerMessage = serde_json::from_str(&message)?;
                    msg
                }
                Message::Binary(_) => {
                    log::debug!("BINARY message from server");
                    continue;
                }
                Message::Ping(_) => {
                    log::debug!("PING from server");
                    continue;
                }
                Message::Pong(_) => {
                    log::debug!("PONG from server");
                    continue;
                }
                Message::Close(_) => {
                    log::debug!("CLOSE message request from server");
                    break;
                }
                Message::Frame(_) => {
                    log::debug!("Frame from server");
                    continue;
                }
            };

            if let Some(cb) = &mut self.server_msg_callback {
                (cb)(&message)
            }

            match message {
                ServerMessage::WaitingForAnotherPlayer => {
                    log::debug!("Waiting for other player...")
                }
                ServerMessage::SetInviteCode { code } => {
                    log::debug!("Got invite code: {}", code);
                }
                ServerMessage::InvalidInviteCode => {
                    log::debug!("Got invalid invite code!");
                    return Ok(ClientEndResult::InvalidInviteCode);
                }
                ServerMessage::OpponentConnected => {
                    log::debug!("Opponent connected");
                }
                ServerMessage::QueryBoatsLayout { rules } => {
                    assert_eq!(&rules, &self.requested_rules);
                    log::debug!("Server requested boats layout");
                    socket
                        .send(Message::Text(serde_json::to_string(
                            &ClientMessage::BoatsLayout {
                                layout: self.layout.clone().unwrap_or_else(|| {
                                    BoatsLayout::gen_random_for_rules(&self.requested_rules)
                                        .unwrap()
                                }),
                            },
                        )?))
                        .await?;
                }
                ServerMessage::WaitingForOtherPlayerConfiguration => {
                    log::debug!("Waiting for other player configuration...")
                }
                ServerMessage::OpponentReady => log::debug!("Other player is ready!"),
                ServerMessage::GameStarting => {
                    log::debug!("The game is starting...");
                    remaining_games -= 1;
                }
                ServerMessage::OpponentMustFire { status } => {
                    assert_eq!(status.opponent_map.boats.number_of_boats(), 0);
                    log::debug!("Other player must fire!")
                }
                ServerMessage::RequestFire { status } => {
                    assert_eq!(status.opponent_map.boats.number_of_boats(), 0);

                    let location = status.find_fire_coordinates_for_bot_type(self.play_as_bot_type);
                    log::debug!("Will fire at {:?}", location);
                    socket
                        .send(Message::Text(serde_json::to_string(
                            &ClientMessage::Fire { location },
                        )?))
                        .await?;
                }
                ServerMessage::FireResult { pos, result } => {
                    log::debug!("Strike at {} result: {:?}", pos.human_print(), result)
                }
                ServerMessage::OpponentFireResult { pos, result } => log::debug!(
                    "Opponent trike at {} result: {:?}",
                    pos.human_print(),
                    result
                ),
                ServerMessage::LostGame { status } => {
                    number_defeats += 1;

                    log::debug!("We lost game :(");
                    log::debug!("Opponent map:\n{}", status.get_opponent_map());
                    log::debug!("Our map:\n{}\n", status.get_your_map());

                    if remaining_games > 0 {
                        socket
                            .send(Message::Text(serde_json::to_string(
                                &ClientMessage::RequestRematch,
                            )?))
                            .await?;
                    } else {
                        break;
                    }
                }
                ServerMessage::WonGame { status } => {
                    number_victories += 1;

                    log::debug!("We won the game !!!!");
                    log::debug!("Opponent map:\n{}\n", status.get_opponent_map());
                    log::debug!("Our map:\n{}\n", status.get_your_map());

                    if remaining_games > 0 {
                        socket
                            .send(Message::Text(serde_json::to_string(
                                &ClientMessage::RequestRematch,
                            )?))
                            .await?;
                    } else {
                        break;
                    }
                }
                ServerMessage::RejectedBoatsLayout { errors } => {
                    log::warn!("Rejected boat layout: {:?}", errors);
                    return Ok(ClientEndResult::InvalidBoatsLayout);
                }
                ServerMessage::SetOpponentName { name } => log::debug!("Opponent name: {}", name),

                ServerMessage::OpponentRequestedRematch => {
                    log::debug!("Opponent requested rematch.");
                    socket
                        .send(Message::Text(serde_json::to_string(
                            &ClientMessage::AcceptRematch,
                        )?))
                        .await?;
                }
                ServerMessage::OpponentAcceptedRematch => {
                    log::debug!("Opponent accepted rematch");
                }
                ServerMessage::OpponentRejectedRematch => {
                    log::debug!("Opponent rejected rematch");
                    return Ok(ClientEndResult::OpponentRejectedRematch);
                }
                ServerMessage::OpponentLeftGame => {
                    log::debug!("Opponent left game");
                    return Ok(ClientEndResult::OpponentLeftGame);
                }
                ServerMessage::OpponentReplacedByBot => {
                    log::debug!("Opponent replaced by bot");
                }
            }
        }

        Ok(ClientEndResult::Finished {
            number_victories,
            number_defeats,
        })
    }
}
